if __name__ == "__main__":
    import sys

    sys.path.insert(0, ".")
    from q2rad.q2rad import main

    main()

from q2gui.q2model import Q2CursorModel
from q2rad.q2raddb import Q2Cursor, SeqMover, insert
from q2gui.q2dialogs import q2AskYN

from q2rad import Q2Form

from q2gui import q2app


import gettext

_ = gettext.gettext

SQL_DATATYPES = (
    "char",
    "varchar",
    "text",
    "longtext",
    "num",
    "dec",
    "decimal",
    "date",
    "time",
    "datetime",
    "integer",
    "bigint",
    "int",
)
HAS_DATADEC = ("dec", "numeric", "num")
HAS_DATALEN = ("char", "varchar") + HAS_DATADEC
WIDGETS = ("line", "text", "code", "button", "check", "radio", "combo", "list", "spin")
# "date;"
# "frame;"
# "grid;"
# "label;"
# "lookup;"
# "progressbar;"
# "relation;"
# "space;"
# "tab;"
# "toolbar;"
# "toolbutton;"

HAS_PIC = "radio;" "combo;" "list;"


class Q2Lines(Q2Form, SeqMover):
    def __init__(self):
        super().__init__("Lines")
        self.no_view_action = True

    def on_init(self):
        self.create_form()
        self.db = q2app.q2_app.db_logic
        cursor: Q2Cursor = self.db.table(table_name="lines", order="seq")
        model = Q2CursorModel(cursor)
        self.set_model(model)

        self.add_action("/crud")
        self.add_seq_actions()

        self.add_action("Run", self.form_runner, hotkey="F4")
        self.add_action("Fill", self.filler)

    def create_form(self):
        self.add_control("id", "", datatype="int", pk="*", ai="*", noform=1, nogrid=1)
        self.add_control(
            "name",
            _("Form"),
            disabled="*",
            to_table="forms",
            to_column="name",
            related="title",
            nogrid=1,
            noform=1,
        )
        self.add_control("column", _("Column name"))
        self.add_control("/")
        if self.add_control("/t", _("Main")):
            self.add_control("/f")
            self.add_control("label", _("Form label"), datatype="char", datalen=100)
            self.add_control("gridlabel", _("Grid label"))
            if self.add_control("/h"):
                self.add_control("seq", _("Sequence number"), datatype="int")
                self.add_control("nogrid", _("No grid"), control="check")
                self.add_control("noform", _("No form"), control="check")
                self.add_control("check", _("Has checkbox"), control="check")
                self.add_control("/s")
                self.add_control("/")
            if self.add_control("/h", _("Control type")):

                self.add_control(
                    "control",
                    gridlabel=_("Control type"),
                    pic=";".join(WIDGETS),
                    control="combo",
                    valid=self.control_valid,
                    datatype="char",
                    datalen=15,
                )
                self.add_control("pic", _("Control data"), datatype="char", datalen=250)
                self.add_control("/")

            if self.add_control("/h", _("Data type")):

                self.add_control(
                    "datatype",
                    gridlabel=_("Data type"),
                    pic=";".join(SQL_DATATYPES),
                    control="combo",
                    valid=self.datatype_valid,
                    datatype="char",
                    datalen=15,
                )
                self.add_control("datalen", _("Data lenght"), datatype="int")
                self.add_control("datadec", _("Decimal lenght"), datatype="int")
                self.add_control("/s")
                self.add_control("/")

            if self.add_control("/h"):  # Db

                self.add_control(
                    "migrate",
                    _("Migrate to database"),
                    control="check",
                    valid=self.database_valid,
                )
                self.add_control(
                    "pk",
                    _("Primary key"),
                    control="check",
                    valid=self.database_valid,
                )
                self.add_control(
                    "ai",
                    _("Autoincrement"),
                    control="check",
                    valid=self.database_valid,
                )
                self.add_control("/s")
                self.add_control("/")
            self.add_control("/")  # Linked
            if self.add_control("/f", _("Linked")):
                self.add_control(
                    "to_table", _("To table"), datatype="char", datalen=100
                )
                self.add_control(
                    "to_column", _("To field"), datatype="char", datalen=100
                )
                self.add_control(
                    "related", _("Data to show"), datatype="char", datalen=100
                )
                self.add_control(
                    "to_form", _("Form to open"), datatype="char", datalen=100
                )

                self.add_control("/")
            self.add_control("/s")

        self.add_control("/t", _("Comment"))
        self.add_control("comment", gridlabel=_("Comments"), datatype="bigtext")
        self.add_control("/t", _("Script When"))
        self.add_control(
            "code_when",
            _("Script When"),
            control="code",
            nogrid="*",
            datatype="bigtext",
        )
        self.add_control("/t", _("Script Show"))
        self.add_control(
            "code_show",
            _("Script When"),
            control="code",
            nogrid="*",
            datatype="bigtext",
        )
        self.add_control("/t", _("Script Valid"))
        self.add_control(
            "code_valid",
            _("Script When"),
            control="code",
            nogrid="*",
            datatype="bigtext",
        )

    def filler(self):
        if self.model.row_count() > 0:
            if q2AskYN("Lines list is not empty! Are you sure") != 2:
                return

        cols = self.q2_app.db_data.db_schema.get_schema_columns(
            self.prev_form.r.form_table
        )
        for x in cols:
            insert(
                "lines",
                {
                    "name": self.prev_form.r.name,
                    "column": x,
                    "label": x,
                    "datatype": cols[x]["datatype"],
                    "datalen": cols[x]["datalen"],
                    "pk": cols[x]["pk"],
                    "ai": cols[x]["ai"],
                    "migrate": "*",
                },
                self.db,
            )
        self.refresh()

    def before_crud_save(self):
        if not self.s.migrate:
            self.s.pk = ""
            self.s.ai = ""

    def form_runner(self):
        self.prev_form.run_action("Run")

    def before_form_show(self):
        self.datatype_valid()
        self.control_valid()
        self.database_valid()
        self.next_sequense()

    def datatype_valid(self):
        self.w.datalen.set_enabled(self.s.datatype in ";".join(HAS_DATALEN))
        self.w.datadec.set_enabled(self.s.datatype in ";".join(HAS_DATADEC))

    def control_valid(self):
        self.w.pic.set_enabled(self.s.control in HAS_PIC)

    def database_valid(self):
        self.w.pk.set_enabled(self.s.migrate)
        self.w.ai.set_enabled(self.s.migrate)
        name = self.prev_form.r.name
        id = self.r.id
        id_where = ""
        if self.crud_mode in ("EDIT", "VIEW"):
            id_where = f" and id <> {id}"
        sql = f"select * from lines where pk='*' and name = '{name}' {id_where}"
        if not self.s.migrate or self.db.cursor(sql).row_count() > 0:
            self.w.pk.set_disabled()
            self.w.ai.set_disabled()



import sys
from xmlrpc.client import Marshaller

if __name__ == "__main__":

    sys.path.insert(0, ".")


from q2rad import Q2App, Q2Form
from q2gui.q2dialogs import q2Mess, q2AskYN, q2WaitShow
from q2gui.q2model import Q2CursorModel
from q2db.schema import Q2DbSchema
from q2db.db import Q2Db
from q2rad.q2actions import Q2Actions
from q2db.cursor import Q2Cursor
from q2rad.q2raddb import read_url, open_url

from q2gui import q2app
from q2rad.q2raddb import q2cursor
from q2rad.q2appmanager import AppManager

# from random import randint

from q2rad.q2appselector import Q2AppSelect
from q2rad.q2modules import Q2Modules
from q2rad.q2forms import Q2Forms
from q2rad.q2lines import Q2Lines
from q2rad.q2market import Q2Market
from q2rad.q2constants import Q2Constants, q2const
from q2rad.q2queries import Q2Queries
from q2rad.q2reports import Q2Reports, Q2RadReport

import traceback
import gettext

# import urllib.request
import os
import json
import subprocess
import shutil


q2_modules = ("q2rad", "q2gui", "q2db", "q2report")
const = q2const()
_ = gettext.gettext


def get_report(report_name):
    content = q2app.q2_app.db_logic.get("reports", f"name='{report_name}'", "content")
    if content:
        return Q2RadReport(content)
    else:
        q2Mess(f"Report not fount: {report_name}")
        return None


def run_module(module_name):
    q2app.q2_app.run_module(module_name)


class Q2RadApp(Q2App):
    def __init__(self, title=""):
        super().__init__(title)
        self.db = None

        self.db_data = None
        self.db_logic = None
        self.dev_mode = False
        self.selected_application = {}
        self.clear_app_info()

        self.q2market_path = "../q2market"

        self.q2market_url = (
            "https://raw.githubusercontent.com/AndreiPuchko/q2market/main/"
        )

        self.assets_url = (
            "https://raw.githubusercontent.com/AndreiPuchko/q2gui/main/assets/"
        )

        qss_file = "q2gui.qss"
        if not os.path.isfile(qss_file):
            qss_url = (
                "https://raw.githubusercontent.com/AndreiPuchko/q2rad/main/q2gui.qss"
            )
            open(qss_file, "w").write(read_url(qss_url).decode("utf-8"))

        if os.path.isfile(qss_file):
            self.style_file = qss_file
            self.set_style_sheet()

        self.set_icon("assets/q2rad.ico")

        self.const = const

    def clear_app_info(self):
        self.app_url = None
        self.app_title = ""
        self.app_version = ""
        self.app_description = ""

    def make_desktop_shortcut(self):
        if "win" in sys.platform:
            subprocess.check_call(["cscript", "make_shortcut.vbs"], shell=True)
        else:
            basepath = os.path.abspath(".")
            desktop_entry = [
                "[Desktop Entry]\n"
                "Name=q2RAD\n"
                f"Exec={basepath}/q2rad/bin/q2rad\n"
                f"Path={basepath}\n"
                f"Icon={basepath}/assets/q2rad.ico\n"
                "Terminal=false\n"
                "Type=Application\n"
            ]

            desktop = os.path.join(os.path.join(os.path.expanduser("~")), "Desktop")
            open(f"{desktop}/q2rad.desktop", "w").writelines("\n".join(desktop_entry))

    def on_start(self):
        if not os.path.isfile("poetry.lock"):
            self.load_assets()
            self.check_q2_update()
        self.open_application(autoload_enabled=True)

    def open_application(self, autoload_enabled=False):
        Q2AppSelect().run(autoload_enabled)
        if self.selected_application != {}:
            self.open_selected_app()
            self.check_app_update()
        else:
            self.close()

    def open_selected_app(self):
        self.clear_app_info()
        self.migrate_db_logic()
        self.migrate_db_data()
        self.run_module("autorun")
        self.set_title(f"{self.app_title}({self.selected_application.get('name', '')})")
        # DEBUG
        # self.run_forms()
        # self.run_queries()
        # self.run_modules()
        # self.run_reports()
        # self.run_app_manager()
        if (
            max(
                [
                    self.db_logic.table("forms").row_count(),
                    self.db_logic.table("lines").row_count(),
                    self.db_logic.table("actions").row_count(),
                    self.db_logic.table("reports").row_count(),
                    self.db_logic.table("modules").row_count(),
                    self.db_logic.table("queries").row_count(),
                ]
            )
            <= 0
        ):
            if (
                q2AskYN("Application is empty! Would you like to download some App?")
                == 2
            ):
                Q2Market().run()

    def migrate_db_data(self):
        data_schema = Q2DbSchema()
        cu = q2cursor(
            """
                select
                    forms.form_table as `table`
                    , lines.column
                    , lines.datatype
                    , lines.datalen
                    , lines.datadec
                    , lines.to_table
                    , lines.to_column
                    , lines.related
                    , lines.ai
                    , lines.pk
                from lines, forms
                where forms.name = lines.name
                    and form_table <>'' and migrate <>''
                order by forms.seq, lines.seq, forms.name
                """,
            self.db_logic,
        )
        for column in cu.records():
            data_schema.add(**column)
        for form in (Q2Constants(),):
            for x in form.get_table_schema():
                data_schema.add(**x)

        self.db_data.set_schema(data_schema)
        self.create_menu()

    def migrate_db_logic(self):
        data_schema = Q2DbSchema()
        for form in (
            Q2Modules(),
            Q2Forms(),
            Q2Lines(),
            Q2Queries(),
            Q2Actions(),
            Q2Reports(),
        ):
            for x in form.get_table_schema():
                data_schema.add(**x)
        self.db_logic.set_schema(data_schema)

    def open_databases(self):
        self.db_data = Q2Db(database_name=self.selected_application["database_data"])
        self.db_logic = Q2Db(database_name=self.selected_application["database_logic"])

    def create_menu(self):
        self.clear_menu()
        self.add_menu("File|About", self.about, icon="info.png")
        self.add_menu("File|Manage", self.run_app_manager, icon="tools.png")
        self.add_menu("File|Constants", self.run_constants)
        self.add_menu("File|-")
        self.add_menu("File|Open", self.open_application, icon="open.png")
        self.add_menu("File|-")
        self.add_menu("File|Close", self.close, toolbar=1, icon="exit.png")

        self.create_form_menu()

        self.dev_mode = self.selected_application.get("dev_mode")

        if self.dev_mode:
            self.add_menu("Dev|Forms", self.run_forms, toolbar=self.dev_mode)
            self.add_menu("Dev|Modules", self.run_modules, toolbar=self.dev_mode)
            self.add_menu("Dev|Querys", self.run_queries, toolbar=self.dev_mode)
            self.add_menu("Dev|Reports", self.run_reports, toolbar=self.dev_mode)
        self.build_menu()
        # self.show_toolbar(False)

    def about(self, text=""):
        about = []
        if self.app_title:
            about.append(f"<b><font size=+1>{self.app_title}</font></b>")
        if self.app_description:
            about.append(f"<i>{self.app_description}</i>")
        if self.app_version:
            about.append(f"Uploaded: {self.app_version}")
        if self.app_url:
            about.append(f"URL: <u>{self.app_url}</u>")
        about.append("")
        if text:
            about.append(text)
        about.append("<b>q2RAD</b>")
        about.append("Versions:")
        about.append(f"<b>Python</b>: {sys.version}<p>")
        w = q2WaitShow(len(q2_modules))
        for package in q2_modules:
            w.step()
            latest_version, current_version = self.get_package_versions(package)
            about.append(
                f"<b>{package}</b>: {current_version}"
                f"{'(' + latest_version + ' avaiable)' if current_version!=latest_version else ''}"
            )
        w.close()
        q2Mess("<br>".join(about))

    def asset_file_loader(self, name):
        asset_url = f"{self.assets_url}/{name}"
        try:
            asset_content = read_url(asset_url)
        except Exception:
            print(f"Error reading {asset_url}")
            return
        try:
            open(f"assets/{name}", "wb").write(asset_content)
        except Exception:
            print(f"Error writing asset/{name}")

    def write_restore_file(self, name, content):
        if "win" in sys.platform:
            u_file = open(f"{name}.bat", "w")
        else:
            u_file = open(os.open(f"{name}.sh", os.O_CREAT | os.O_WRONLY, 0o777), "w")
        u_file.write(content)
        u_file.close()

    def load_assets(self, force_reload=False):
        if os.path.isdir("assets") and force_reload is False:
            return
        if not os.path.isdir("assets"):
            os.mkdir("assets")
        # first run
        # load icons
        icons = [getattr(q2app, x) for x in dir(q2app) if x.endswith("ICON")]
        icons.append("q2gui.ico")

        w = q2WaitShow(len(icons))
        for x in icons:
            w.step(x)
            self.asset_file_loader(x)
        w.close()
        if os.path.isfile("assets/q2gui.ico"):
            shutil.copyfile("assets/q2gui.ico", "assets/q2rad.ico")

        self.set_icon("assets/q2rad.ico")

        if os.path.isfile("poetry.lock"):
            return

        # create update_q2rad.sh
        self.write_restore_file(
            "update_q2rad",
            ("" if "win" in sys.platform else "#!/bin/bash\n")
            + (
                "q2rad\\scripts\\activate "
                if "win" in sys.platform
                else "source q2rad/bin/activate"
            )
            + "&& pip install --upgrade q2gui"
            + "&& pip install --upgrade q2db"
            + "&& pip install --upgrade q2report"
            + "&& pip install --upgrade q2rad",
        )

        # create run_q2rad
        self.write_restore_file(
            "run_q2rad",
            ("" if "win" in sys.platform else "#!/bin/bash\n")
            + (
                "start q2rad\\scripts\\pythonw.exe -m q2rad"
                if "win" in sys.platform
                else "q2rad/bin/q2rad\n"
            ),
        )
        if "win" in sys.platform:
            open("run_q2rad.vbs", "w").write(
                'WScript.CreateObject("WScript.Shell").Run '
                '"q2rad\\scripts\\pythonw.exe -m q2rad", 0, false'
            )

            open("make_shortcut.vbs", "w").write(
                'Set oWS = WScript.CreateObject("WScript.Shell")\n'
                'Set oLink = oWS.CreateShortcut(oWS.SpecialFolders("Desktop") & "\\q2RAD.lnk")\n'
                'cu = WScript.CreateObject("Scripting.FileSystemObject").'
                "GetParentFolderName(WScript.ScriptFullName)\n"
                'oLink.TargetPath = cu & "\\run_q2rad.vbs"\n'
                'oLink.WorkingDirectory = cu & ""\n'
                'oLink.Description = "q2RAD"\n'
                'oLink.IconLocation = cu & "\\assets\\q2rad.ico"\n'
                "oLink.Save\n"
            )

        if q2AskYN("Can I make desktop shortcut?") == 2:
            self.make_desktop_shortcut()

    def get_package_versions(self, package):
        latest_version = json.load(
            open_url(f"https://pypi.python.org/pypi/{package}/json")
        )["info"]["version"]
        current_version = sys.modules[package].__version__
        return latest_version, current_version

    def update_packages(self):
        upgraded = []
        w = q2WaitShow(len(q2_modules))
        for package in q2_modules:
            if w.step(package):
                break
            latest_version, current_version = self.get_package_versions(package)
            if latest_version != current_version:
                runpip = lambda: subprocess.check_call(  # noqa:E731
                    [
                        sys.executable.replace("w.exe", ".exe"),
                        "-m",
                        "pip",
                        "install",
                        "--upgrade",
                        "--no-cache-dir",
                        f"{package}=={latest_version}",
                    ],
                    shell=True if "win" in sys.platform else False,
                )
                try:
                    runpip()
                except Exception:
                    try:
                        runpip()
                    except Exception:
                        pass

                latest_version, new_current_version = self.get_package_versions(package)
                upgraded.append(
                    f"{package} - "
                    f"<b>{current_version}</b> => "
                    f"<b>{latest_version}</b>"
                )
        w.close()
        if upgraded:
            mess = (
                "Upgrading complete!<p>" "The program will be restarted!" "<p><p>"
            ) + "<p>".join(upgraded)
        else:
            mess = "Updates not found!<p>"
        q2Mess(mess)
        if upgraded:
            os.execv(sys.executable, [sys.executable, "-m", "q2rad"])
            # if "win" in sys.platform:
            #     os.execv(sys.executable, [sys.executable, "-m", "q2rad"])
            # else:
            #     os.execv(sys.argv[0], sys.argv)
            # self.close()
        pass

    def check_app_update(self):
        if self.app_url and self.app_version:
            market_version = read_url(self.app_url + ".version").decode("utf-8")
            if market_version != self.app_version:
                if q2AskYN(
                    f"Update for App <b>{self.app_title}</b> detected!"
                    f"<p>Current version <b>{self.app_version}</b>"
                    f"<p>New version <b>{market_version}</b>"
                    "<p>Download and install?"
                ):
                    data = json.load(open_url(self.app_url + ".json"))
                    AppManager.import_json_app(data)
                    self.open_selected_app()

    def check_q2_update(self):
        can_upgrade = False
        for package in q2_modules:
            latest_version, current_version = self.get_package_versions(package)
            can_upgrade = latest_version != current_version
            if can_upgrade:
                break
        if can_upgrade:
            if (
                q2AskYN(
                    "Updates for q2* packages are avaiable!<p><p>"
                    "Do you want to proceed with update?<p><p>"
                    "The program will be restarted after the update!"
                )
                == 2
            ):
                self.update_packages()

    def run_constants(self):
        Q2Constants().run()

    def run_app_manager(self):
        AppManager().run()

    def create_form_menu(self):
        cu = q2cursor(
            """select
                menu_path
                , title
                , toolbar
                , menu_text
                , menu_before
                , menu_tiptext
                , menu_separator
                , name
            from forms
            where menu_path <> ''
            order by seq
            """,
            self.db_logic,
        )
        for x in cu.records():
            menu_path = (
                x["menu_path"]
                + "|"
                + (x["menu_text"] if x["menu_text"] else x["title"])
            )

            def menu_worker(name):
                def real_worker():
                    self.run_form(name)

                return real_worker

            self.add_menu(
                menu_path,
                worker=menu_worker(x["name"]),
                toolbar=x["toolbar"],
                before=x["menu_before"],
            )

    def run_forms(self):
        Q2Forms().run()

    def run_modules(self):
        Q2Modules().run()

    def run_queries(self):
        Q2Queries().run()

    def run_reports(self):
        Q2Reports().run()

    def run_form(self, name):
        self.get_form(name).run()

    def get_form(self, name):
        if not name:
            return
        form_dic = self.db_logic.get("forms", f"name ='{name}'")

        sql = f"""
            select
                column
                ,label
                ,gridlabel
                ,nogrid
                ,noform
                ,`check`
                ,control
                ,pic
                ,datatype
                ,datalen
                ,datadec
                ,migrate
                ,pk
                ,ai
                ,to_table
                ,to_column
                ,related
                ,to_form
                ,code_valid as valid
                /*,code_when as when
                ,code_show as show*/
            from lines
            where name = '{name}'
            order by seq
            """
        cu = q2cursor(sql, self.db_logic)

        form = Q2Form(form_dic["title"])
        form.no_view_action = 1
        form.ok_button = form_dic["ok_button"]
        form.cancel_button = form_dic["cancel_button"]

        form.valid = self.code_runner(form_dic["form_valid"], form)

        form.before_form_build = self.code_runner(form_dic["before_form_build"], form)
        form.before_grid_build = self.code_runner(form_dic["before_grid_build"], form)

        form.before_form_show = self.code_runner(form_dic["before_form_show"], form)
        form.after_form_show = self.code_runner(form_dic["after_form_show"], form)

        form.before_grid_show = self.code_runner(form_dic["before_grid_show"], form)
        form.after_grid_show = self.code_runner(form_dic["after_grid_show"], form)

        form.before_crud_save = self.code_runner(form_dic["before_crud_save"], form)
        form.after_crud_save = self.code_runner(form_dic["after_crud_save"], form)

        form.before_delete = self.code_runner(form_dic["before_delete"], form)
        form.after_delete = self.code_runner(form_dic["after_delete"], form)

        # add controls
        for x in cu.records():
            if x.get("to_form"):
                x["to_form"] = self.get_form(x["to_form"])
            x["valid"] = self.code_runner(x["valid"], form)
            # x["show"] = self.code_runner(x["show"], form)
            # x["when"] = self.code_runner(x["when"], form)
            form.add_control(**x)

        # add actions
        if form_dic["form_table"]:
            form_cursor: Q2Cursor = self.db_data.table(
                table_name=form_dic["form_table"]
            )
            form_model = Q2CursorModel(form_cursor)
            form.set_model(form_model)
            sql = f"select * from actions where name = '{name}' order by seq"
            cu = q2cursor(sql, self.db_logic)
            for x in cu.records():
                if x["action_mode"] == "1":
                    form.add_action("/crud")
                elif x["action_mode"] == "3":
                    form.add_action("-")
                else:
                    if x["child_form"] and x["child_where"]:
                        child_form_name = x["child_form"]
                        form.add_action(
                            x["action_text"],
                            self.code_runner(x["action_worker"])
                            if x["action_worker"]
                            else None,
                            child_form=lambda: self.get_form(child_form_name),
                            child_where=x["child_where"],
                            hotkey=x["action_key"],
                            eof_disabled=1
                        )
                    else:
                        form.add_action(
                            x["action_text"],
                            self.code_runner(x["action_worker"], form=form)
                            if x["action_worker"]
                            else None,
                            hotkey=x["action_key"],
                            eof_disabled=x["eof_disabled"],
                        )
        return form

    def code_error(self):
        return traceback.format_exc().replace("\n", "<br>").replace(" ", "&nbsp;")

    def code_compiler(self, script):
        script = script.replace("return", "pass;")
        try:
            code = compile(script, "'<worker>'", "exec")
            return {"code": code, "error": ""}
        except Exception:
            trace = self.code_error()
            return {
                "code": False,
                "error": f"""Compile error:<br><br>{trace}""",
            }

    def code_runner(self, script, form=None, __name__="__main__"):
        _form = form
        # to provide return ability for exec
        if "return " in script:
            nsl = []
            for x in script.split("\n"):
                if "return " in x:
                    if x.strip() == "return":
                        x = x.replace("return", "raise ReturnEvent")
                    else:
                        x = x.replace("return", "RETURN = ") + "; raise ReturnEvent"
                nsl.append(x)
            script = "\n".join(nsl)

        def real_runner():
            # make exec stop on return
            class ReturnEvent(Exception):
                pass

            __locals_dict = {
                "RETURN": None,
                "ReturnEvent": ReturnEvent,
                "mem": _form,
                "form": _form,
                "self": self,
                "q2_app": self,
                "myapp": self,
                "__name__": __name__,
            }
            code = self.code_compiler(script)
            if code["code"]:
                try:
                    exec(code["code"], globals(), __locals_dict)
                except ReturnEvent:
                    pass
                except Exception:
                    trace = self.code_error()
                    q2Mess(f"""Runtime error:<br><br>{trace}""")
            else:
                q2Mess(code["error"])

            return __locals_dict["RETURN"]

        return real_runner

    def run_module(self, name=""):
        script = self.db_logic.get("modules", f"name = '{name}'", "script")
        if script:
            return self.code_runner(script)()


def main():
    app = Q2RadApp("q2RAD")
    app.dev_mode = 1
    app.run()


if __name__ == "__main__":
    main()



if __name__ == "__main__":
    import sys

    sys.path.insert(0, ".")
    from q2rad.q2rad import main

    main()

from q2db.cursor import Q2Cursor
from q2gui.q2model import Q2CursorModel
from q2rad.q2raddb import  q2cursor, SeqMover
from q2gui import q2app
from q2rad import Q2Form

import gettext

_ = gettext.gettext


class Q2Actions(Q2Form, SeqMover):
    def __init__(self):
        super().__init__("Actions")
        self.no_view_action = True

    def on_init(self):
        self.db = q2app.q2_app.db_logic
        self.create_form()

        cursor: Q2Cursor = self.db.table(table_name="actions", order="seq")
        model = Q2CursorModel(cursor)
        self.set_model(model)

        self.add_action("/crud")
        self.add_seq_actions()
        self.add_action("Run", self.form_runner, hotkey="F4")

    def create_form(self):
        self.add_control("id", "", datatype="int", pk="*", ai="*", noform=1, nogrid=1)
        self.add_control(
            "name",
            _("Form"),
            disabled="*",
            to_table="forms",
            to_column="name",
            related="title",
            nogrid=1,
            noform=1,
        )
        self.add_control("action_text", _("Action text"))
        self.add_control("/")
        if self.add_control("/t", _("Main")):
            if self.add_control("/f"):
                self.add_control("seq", _("Sequence number"), datatype="int")

                self.add_control(
                    "action_mode",
                    _("Action mode"),
                    pic="CRUD actions;Single Action;Separator",
                    datatype="int",
                    control="radio",
                    valid=self.action_mode_valid,
                )
                self.add_control(
                    "action_mess", _("Action message"), datatype="char", datalen=100
                )
                self.add_control(
                    "action_icon", _("Action icon"), datatype="char", datalen=100
                )
                self.add_control(
                    "action_key", _("Hot key"), datatype="char", datalen=10
                )
                self.add_control(
                    "eof_disabled",
                    _("Disabled for empty grid"),
                    control="check",
                    datatype="char",
                    datalen=1,
                )
                self.add_control("/")

            if self.add_control("/f", "Child grid"):
                self.add_control(
                    "child_form", _("Child form"), datatype="char", datalen=100
                )
                self.add_control(
                    "child_where", _("Child field"), datatype="char", datalen=100
                )
                self.add_control(
                    "child_noshow",
                    _("Don't show"),
                    control="check",
                    datatype="char",
                    datalen=1,
                )
                self.add_control(
                    "child_copy_mode",
                    _("Copy mode"),
                    pic=_("Ask;Always;Newer"),
                    control="radio",
                    datatype="int",
                )
                self.add_control("/")

            self.add_control("/s")
        self.add_control("/t", _("Action Script"))
        self.add_control(
            "action_worker",
            gridlabel=_("Action Script"),
            datatype="bigtext",
            control="code",
            nogrid="*",
        )
        self.add_control("/t", _("Comment"))
        self.add_control(
            "comment", gridlabel=_("Comments"), datatype="bigtext", control="text"
        )

    def form_runner(self):
        self.prev_form.run_action("Run")

    def action_mode_valid(self):
        for x in self.widgets():
            if x.startswith("_"):
                continue
            elif x.startswith("/"):
                continue
            elif not hasattr(self.widgets()[x], "set_disabled"):
                continue
            elif x in ("action_mode", "ordnum", "comment", "seq"):
                continue
            else:
                self.widgets()[x].set_disabled(self.s.action_mode != "2")

    def before_form_show(self):
        self.action_mode_valid()
        self.next_sequense()



if __name__ == "__main__":
    import sys

    sys.path.insert(0, ".")
    from q2rad.q2rad import main

    main()


from q2db.cursor import Q2Cursor
from q2db.schema import Q2DbSchema

from q2rad.q2raddb import q2cursor, SeqMover
from q2gui.q2model import Q2CursorModel
from q2gui.q2model import Q2Model

# from q2gui.q2utils import int_

from q2gui.q2dialogs import q2Mess
from q2gui import q2app

from q2rad.q2lines import Q2Lines
from q2rad.q2actions import Q2Actions

from q2rad import Q2Form

import gettext

_ = gettext.gettext


class Q2Forms(Q2Form, SeqMover):
    def __init__(self):
        super().__init__("Forms")
        self.no_view_action = True

    def on_init(self):
        self.create_form()
        self.db = q2app.q2_app.db_logic
        cursor: Q2Cursor = self.db.table(table_name="forms", order="seq")
        model = Q2CursorModel(cursor)
        model.set_order("seq").refresh()
        self.set_model(model)

        self.add_action("/crud")

        self.add_action(
            "Lines",
            child_form=Q2Lines,
            child_where="name='{name}'",
            hotkey="F2",
            eof_disabled=1,
        )
        self.add_action(
            "Actions",
            child_form=Q2Actions,
            child_where="name='{name}'",
            hotkey="F3",
            eof_disabled=1,
        )

        self.add_seq_actions()

        self.add_action("Migrate to DB", self.q2_app.migrate_db_data, eof_disabled=1)
        self.add_action("Run", self.form_runner, hotkey="F4", eof_disabled=1)

    def create_form(self):
        self.add_control("name", _("Name"), datatype="char", datalen=100, pk="*")
        self.add_control("/")

        if self.add_control("/t", _("Main")):
            self.add_control("/h")
            self.add_control("title", _("Title"), datatype="char", datalen=100)
            self.add_control("seq", _("Sequence number"), datatype="int")
            self.add_control("/")
            self.add_control("/f", _("Main menu"))
            self.add_control(
                "menu_path", _("Menu bar path"), datatype="char", datalen=100
            )
            self.add_control("menu_text", _("Menu text"), datatype="char", datalen=100)
            self.add_control(
                "menu_before", _("Before path"), datatype="char", datalen=100
            )
            self.add_control(
                "menu_tiptext", _("Tip text"), datatype="char", datalen=100
            )
            if self.add_control("/h"):
                self.add_control(
                    "menu_separator",
                    _("Add separator before"),
                    control="check",
                    datatype="char",
                    datalen=1,
                )
                self.add_control(
                    "toolbar",
                    _("Show in app toolbar"),
                    control="check",
                    datatype="char",
                    datalen=1,
                )

                self.add_control("/")
            self.add_control("/")
            self.add_control("/h")
            self.add_control(
                "ok_button",
                _("Add OK button"),
                datatype="char",
                datalen=1,
                control="check",
            )
            self.add_control(
                "cancel_button",
                _("Add Cancel button"),
                datatype="char",
                datalen=1,
                control="check",
            )
            self.add_control("/")

            self.add_control("/f", _("Data"))
            if self.add_control("/h", _("Data table")):
                self.add_control(
                    "select_data_storage_file",
                    _("?"),
                    mess=_("Open list of existing tables"),
                    control="button",
                    datalen=3,
                    valid=self.select_data_storage_table,
                )
                self.add_control("form_table", gridlabel=_("Table"))
                self.add_control("/")
            if self.add_control("/h", _("Sort by")):
                self.add_control(
                    "select_data_sort_column",
                    _("?"),
                    mess=_("Open column list"),
                    control="button",
                    datalen=3,
                    valid=self.select_table_sort_column,
                )
                self.add_control("form_table_sort", "", datatype="char", datalen=100)
                self.add_control("/")
            self.add_control("/")
            self.add_control("/s")

        self.add_control("/t", _("Comment"))
        self.add_control("/f")
        self.add_control("comment", gridlabel=_("Comments"), datatype="bigtext")

        if self.add_control("/t", _("Build")):
            self.add_control("/vs")
            self.add_control("/v")
            self.add_control(
                "before_form_build",
                label=_("Before Form Build"),
                nogrid="*",
                control="code",
            )
            self.add_control("/")
            self.add_control("/v")
            self.add_control(
                "before_grid_build",
                label=_("Before Grid Build"),
                nogrid="*",
                control="code",
            )
            self.add_control("/")

        if self.add_control("/t", _("Grid")):
            self.add_control("/vs")
            self.add_control("/v")
            self.add_control(
                "before_grid_show",
                label=_("Before Grid Show"),
                nogrid="*",
                control="code",
            )
            self.add_control("/")
            self.add_control("/v")
            self.add_control(
                "after_grid_show",
                label=_("After Grid Show"),
                nogrid="*",
                control="code",
            )
            self.add_control("/")

        if self.add_control("/t", _("Form")):
            self.add_control("/vs")
            self.add_control("/v")
            self.add_control(
                "before_form_show",
                label=_("Before Form Show"),
                nogrid="*",
                control="code",
            )
            self.add_control("/")
            self.add_control("/v")
            self.add_control(
                "after_form_show",
                label=_("After Form Show"),
                nogrid="*",
                control="code",
            )
            self.add_control("/")

        if self.add_control("/t", _("Save")):
            self.add_control("/vs")
            self.add_control("/v")
            self.add_control(
                "before_crud_save", label=_("Before save"), nogrid="*", control="code"
            )
            self.add_control("/")
            self.add_control("/v")
            self.add_control(
                "after_crud_save", label=_("After save"), nogrid="*", control="code"
            )
            self.add_control("/")

        if self.add_control("/t", _("Delete")):
            self.add_control("/vs")
            self.add_control("/v")
            self.add_control(
                "before_delete", label=_("Before delete"), nogrid="*", control="code"
            )
            self.add_control("/")
            self.add_control("/v")
            self.add_control(
                "after_delete", label=_("After delete"), nogrid="*", control="code"
            )
            self.add_control("/")

        if self.add_control("/t", _("Valid")):
            self.add_control(
                "form_valid",
                label=_(""),
                nogrid="*",
                control="code",
            )

    def select_data_storage_table(self):
        setta = Q2Form("Select table")
        setta.add_control("table", "Table")
        setta.no_view_action = 1
        model = Q2Model()
        model.set_records(
            [{"table": x} for x in self.q2_app.db_data.db_schema.get_schema_tables()]
        )
        setta.set_model(model)
        setta.heap.selected_table = None

        def select_table():
            setta.heap.selected_table = setta.r.table
            setta.close()

        setta.add_action(
            _("Select"),
            select_table,
            hotkey="Enter",
            tag="select",
            eof_disabled=1,
        )
        setta.run()
        self.s.form_table = setta.heap.selected_table
        if self.s.name == "":
            self.s.name = self.s.form_table
        if self.s.title == "":
            self.s.title = self.s.form_table

    def select_table_sort_column(self):
        setta = Q2Form("Select column")
        setta.add_control("col", "Table")
        setta.no_view_action = 1
        model = Q2Model()
        model.set_records(
            [
                {"col": x}
                for x in self.q2_app.db_data.db_schema.get_schema_columns(
                    self.s.form_table
                )
            ]
        )
        setta.set_model(model)
        setta.heap.selected_col = None

        def select_column():
            setta.heap.selected_col = setta.r.col
            setta.close()

        setta.add_action(
            _("Select"),
            select_column,
            hotkey="Enter",
            tag="select",
            eof_disabled=1,
        )
        setta.run()
        if setta.heap.selected_col:
            self.s.form_table_sort += ", " if self.s.form_table_sort else ""
            self.s.form_table_sort += setta.heap.selected_col

    def form_runner(self):
        name = self.r.name
        self.q2_app.run_form(name)

    def before_form_show(self):
        self.next_sequense()

    def before_crud_save(self):
        if self.s.name == "":
            q2Mess(_("Give me some NAME!!!"))
            self.w.name.set_focus()
            return False



if __name__ == "__main__":
    import sys

    sys.path.insert(0, ".")
    from q2rad.q2rad import main

    main()

from q2db.cursor import Q2Cursor
from q2gui.q2model import Q2CursorModel
from q2gui.q2utils import int_, num
from q2gui import q2app
from q2gui.q2dialogs import q2Mess

from q2rad import Q2Form
from q2gui.q2form import NEW, COPY

import urllib.request


def open_url(url):
    return urllib.request.urlopen(url)


def read_url(url):
    return open_url(url).read()


class q2cursor(Q2Cursor):
    def __init__(self, sql="", q2_db=None):
        if q2_db is None:
            q2_db = q2app.q2_app.db_data
        super().__init__(q2_db, sql)

    def browse(self):
        if self.row_count() <= 0:
            q2Mess(
                f"""Query<br>
                        <b>{self.sql}</b><br>
                        returned no records,<br>
                        <font color=red>
                        {self.last_sql_error()}
                    """
            )
        else:
            form = Q2Form(self.sql)
            for x in self.record(0):
                form.add_control(x, x, datalen=250)
            form.set_model(Q2CursorModel(self))
            form.run()
            return self


def insert(table, row, q2_db=None):
    if q2_db is None:
        q2_db = q2app.q2_app.db_data
    return q2_db.insert(table, row)


class SeqMover:
    """create actions for swapping rows in forms, lines, actions"""

    def add_seq_actions(self):
        self.add_action("-")
        self.add_action(
            "Move up", self.move_seq_up, icon="arrow-up.png", eof_disabled=1
        )
        self.add_action(
            "Move down", self.move_seq_down, icon="arrow-down.png", eof_disabled=1
        )
        self.add_action("-")

    def move_seq_up(self):
        if self.current_row > 0:
            nr = self.model.get_record(self.current_row - 1)
            cr = self.model.get_record(self.current_row)
            if nr["seq"] == cr["seq"]:
                nr["seq"] = "%s" % (int_(nr["seq"]) - 1)
            else:
                nr["seq"], cr["seq"] = cr["seq"], nr["seq"]
            self.model.update(nr)
            self.model.update(cr)
            self.refresh()
            self.set_grid_index(self.current_row - 1)

    def move_seq_down(self):
        if self.current_row < self.model.row_count() - 1:
            nr = self.model.get_record(self.current_row + 1)
            cr = self.model.get_record(self.current_row)
            if nr["seq"] == cr["seq"]:
                nr["seq"] = "%s" % (int_(nr["seq"]) + 1)
            else:
                nr["seq"], cr["seq"] = cr["seq"], nr["seq"]
            self.model.update(nr)
            self.model.update(cr)
            self.refresh()
            self.set_grid_index(self.current_row + 1)

    def next_sequense(self):
        if self.crud_mode in (NEW, COPY):
            self.s.seq = self.model.cursor.get_next_sequence("seq", num(self.r.seq))



if __name__ == "__main__":
    import sys

    sys.path.insert(0, ".")
    from q2rad.q2rad import main

    main()

from q2db.cursor import Q2Cursor
from q2gui.q2model import Q2CursorModel
from q2gui.q2dialogs import q2AskYN
from q2rad import Q2Form
from q2gui import q2app
import gettext

from q2gui.q2app import Q2Actions

_ = gettext.gettext


class Q2Constants(Q2Form):
    def __init__(self):
        super().__init__("Constants")
        self.no_view_action = True

    def on_init(self):
        self.editor_actions = Q2Actions()
        self.add_control("const_name", _("Name"), datatype="char", datalen=100, pk="*")
        self.add_control("const_text", _("Label"), datatype="char", datalen=250)
        self.add_control("const_value", _("Value"), datatype="text")
        self.add_control("comment", _("Comment"), datatype="text")

        cursor: Q2Cursor = self.q2_app.db_data.table(table_name="constants")
        model = Q2CursorModel(cursor)
        model.set_order("const_name").refresh()
        self.set_model(model)
        self.add_action("/crud")


class q2const:
    def __getattr__(self, __name):
        return q2app.q2_app.db_data.get(
            "constants", f"const_name = '{__name}'", "const_value"
        )

    def __setattr__(self, __name, __value):
        const_name = self.get_const_name(__name)
        if const_name:
            q2app.q2_app.db_data.update(
                "constants", {"const_name": __name, "const_value": __value}
            )
        else:
            q2app.q2_app.db_data.insert(
                "constants", {"const_name": __name, "const_value": __value}
            )

    def get_const_name(self, __name):
        const_name = q2app.q2_app.db_data.get(
            "constants", f"const_name = '{__name}'", "const_name"
        )

        return const_name

    def check(self, const_name="", const_text="", const_value="", comment=""):
        if const_name == "":
            return
        _const_name = self.get_const_name(const_name)
        if not _const_name:
            q2app.q2_app.db_data.insert(
                "constants",
                {
                    "const_name": const_name,
                    "const_value": const_value,
                    "const_text": const_text,
                    "comment": comment,
                },
            )
